#include <math.h>
#include <malloc.h>
#include <string.h>

#include "outfit.h"
#include "physics.h"
#include "main.h"
#include "log.h"
#include "rng.h"
#include "pilot.h"
#include "weapon.h"

#define SIZE_APROX  0.8 // Aproximation for circle collision detection.

// Some stuff from pilot.
extern Pilot** pilot_stack;
extern int pilots;

typedef struct Weapon {
  Solid* solid; // Actually has its own solid. :D

  unsigned int parent; // The pilot that just shot at you!
  const Outfit* outfit; // Related outfit that fired.
  unsigned int timer; // Mainly used to see when the weapon was fired.

  // Update position and render.
  void(*update)(struct Weapon*, const double, WeaponLayer); // Position update and render.
  void(*think)(struct Weapon*); // Some missiles need to be inteligent.
} Weapon;

// Behind Pilot layer.
static Weapon** backLayer = NULL; // Behind pilots.
static int nbackLayer     = 0;    // Number of elements.
static int mbackLayer     = 0;    // Allocated memory size.
// Behind player layer.
static Weapon** frontLayer = NULL;  // Behind pilots.
static int nfrontLayer    = 0;    // Number of elements.
static int mfrontLayer    = 0;    // Allocated memory size.


static Weapon* weapon_create(const Outfit* outfit, const double dir,
      const Vec2* pos, const Vec2* vel, unsigned int parent);

static void weapon_render(const Weapon* w);
static void weapon_update(Weapon* w, const double dt, WeaponLayer layer);
static void weapon_destroy(Weapon* w, WeaponLayer layer);
static void weapon_free(Weapon* w);

// Update all weapons in the layer.
void weapons_update(const double dt, WeaponLayer layer) {
  Weapon** wlayer;
  int nlayer;

  switch(layer) {
    case WEAPON_LAYER_BG:
      wlayer = backLayer;
      nlayer = nbackLayer;
      break;
    case WEAPON_LAYER_FG:
      wlayer = frontLayer;
      nlayer = nfrontLayer;
      break;
  }
  int i;
  for(i = 0; i < nlayer; i++) {
    switch(wlayer[i]->outfit->type) {
      case OUTFIT_TYPE_BOLT:
        if(SDL_GetTicks() > (wlayer[i]->timer + 1000*(unsigned int)
              wlayer[i]->outfit->range/wlayer[i]->outfit->speed)) {
          weapon_destroy(wlayer[i],layer);
          continue;
        }
        break;
      default:
        break;
    }
    weapon_update(wlayer[i], dt, layer);
  }
}

// Render the weapons.
static void weapon_render(const Weapon* w) {
  int sprite;
  gl_texture* t = w->outfit->gfx_space;

  // Get the sprite corresponding to the direction facing.
  sprite = (int)(w->solid->dir / (2.0*M_PI / (t->sy*t->sx)));

  gl_blitSprite(t, &w->solid->pos, sprite % (int)t->sx, sprite / (int)t->sy);
}

// Update the weapon.
static void weapon_update(Weapon* w, const double dt, WeaponLayer layer) {
  int i;
  for(i = 0; i < pilots; i++) {
    if((w->parent != pilot_stack[i]->id) &&
          (DIST(w->solid->pos, pilot_stack[i]->solid->pos) < (SIZE_APROX +
          w->outfit->gfx_space->sw/2. + pilot_stack[i]->ship->gfx_ship->sw/2.))) {
      pilot_hit(pilot_stack[i], w->outfit->damage_shield, w->outfit->damage_armor);
      weapon_destroy(w, layer);
      return;
    }
  }
  if(w->think) (*w->think)(w);
  (*w->solid->update)(w->solid, dt);

  weapon_render(w);
}

static Weapon* weapon_create(const Outfit* outfit, const double dir, const Vec2* pos,
      const Vec2* vel, unsigned int parent) {
  Vec2 v;
  double mass = 1; // Presumer lasers have a mass of 1.
  double rdir = dir; // Real direction (accuracy).
  Weapon* w = MALLOC_L(Weapon);
  w->parent = parent; // Non-Changeable.
  w->outfit = outfit; // Non-Changeable.
  w->update = weapon_update;
  w->timer = SDL_GetTicks();
  w->think = NULL;

  switch(outfit->type) {
    case OUTFIT_TYPE_BOLT:
      rdir += RNG(-outfit->accuracy/2., outfit->accuracy/2.)/180.*M_PI;
      if((rdir > 2.*M_PI) || (rdir < 0.)) rdir = fmod(rdir, 2.*M_PI);
      vect_cset(&v, VX(*vel)+outfit->speed*cos(rdir), VANGLE(*vel)+outfit->speed*sin(rdir));
      w->solid = solid_create(mass, rdir, pos, &v);
      break;
    default:
      break;
  }
  return w;
}

// Add a new weapon.
void weapon_add(const Outfit* outfit, const double dir, const Vec2* pos, const Vec2* vel,
      unsigned int parent, WeaponLayer layer) {
  if(!outfit_isWeapon(outfit)) {
    ERR("Trying to create a weapon from a non-Weapon type Outfit");
    return;
  }

  Weapon* w = weapon_create(outfit, dir, pos, vel, parent);

  // Set the propper layer.
  Weapon** curLayer = NULL;
  int* mLayer = NULL;
  int* nLayer = NULL;
  switch(layer) {
    case WEAPON_LAYER_BG:
      curLayer = backLayer;
      nLayer = &nbackLayer;
      mLayer = &mbackLayer;
      break;
    case WEAPON_LAYER_FG:
      curLayer = frontLayer;
      nLayer = &nfrontLayer;
      mLayer = &mfrontLayer;
      break;
    default:
      ERR("Invalid WEAPON_LAYER specified.");
      return;
  }
  if(*mLayer > *nLayer) // More memory allocated than what we need.
    curLayer[(*nLayer)++] = w;
  else { // Need to allocate more memory.
    switch(layer) {
      case WEAPON_LAYER_BG:
        curLayer = backLayer = realloc(curLayer, (++(*mLayer))*sizeof(Weapon));
        break;
      case WEAPON_LAYER_FG:
        curLayer = frontLayer = realloc(curLayer, (++(*mLayer))*sizeof(Weapon*));
        break;
    }
    curLayer[(*nLayer)++] = w;
  }
}

// Destroy the weapon.
static void weapon_destroy(Weapon* w, WeaponLayer layer) {
  int i;
  Weapon** wlayer;
  int* nlayer;
  switch(layer) {
    case WEAPON_LAYER_BG:
      wlayer = backLayer;
      nlayer = &nbackLayer;
      break;
    case WEAPON_LAYER_FG:
      wlayer = frontLayer;
      nlayer = &nfrontLayer;
      break;
  }
  for(i = 0; wlayer[i] != w; i++); // Get us to the current posision.
  weapon_free(wlayer[i]);
  (*nlayer)--;
  
  for(; i < (*nlayer); i++)
    wlayer[i] = wlayer[i+1];
}

// Clear the weapon.
static void weapon_free(Weapon* w) {
  solid_free(w->solid);
  free(w);
}

// Clear all the weapons, do not free the layers.
void weapon_clear(void) {
  int i;
  for(i = 0; i < nbackLayer; i++)
    weapon_free(backLayer[i]);
  nbackLayer = 0;
  for(i = 0; i < nfrontLayer; i++)
    weapon_free(frontLayer[i]);
  nfrontLayer = 0;
}

void weapon_exit(void) {
  weapon_clear();
  free(backLayer);
  free(frontLayer);
}

